@@ -658,10 +658,8 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 	while ((onevalue = getc_unlocked(f)) != EOF) {
 		char input = (char)onevalue;
 
-		if (control_C_was_pressed) {
-			statusline(ALERT, _("Interrupted"));
+		if (control_C_was_pressed)
 			break;
-		}
 
 		/* When the byte before the current one is a CR and we're doing
 		 * format conversion, then strip this CR when it's before a LF
@@ -735,6 +733,9 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 	if (ferror(f) && errornumber != EINTR && errornumber != 0)
 		statusline(ALERT, strerror(errornumber));
 
+	if (control_C_was_pressed)
+		statusline(ALERT, _("Interrupted"));
+
 	fclose(f);
 
 	if (fd > 0 && !undoable && !ISSET(VIEW_MODE))
@@ -950,7 +951,7 @@ bool execute_command(const char *command)
 	/* Create a pipe to read the command's output from, and, if needed,
 	 * a pipe to feed the command's input through. */
 	if (pipe(from_fd) == -1 || (should_pipe && pipe(to_fd) == -1)) {
-		statusline(ALERT, _("Could not create pipe"));
+		statusline(ALERT, _("Could not create pipe: %s"), strerror(errno));
 		return FALSE;
 	}
 
@@ -986,7 +987,7 @@ bool execute_command(const char *command)
 	close(from_fd[1]);
 
 	if (pid_of_command == -1) {
-		statusline(ALERT, _("Could not fork"));
+		statusline(ALERT, _("Could not fork: %s"), strerror(errno));
 		close(from_fd[0]);
 		return FALSE;
 	}
@@ -1619,14 +1620,20 @@ bool make_backup_of(char *realname)
 		goto problem;
 
 	/* Try to change owner and group to those of the original file;
-	 * ignore errors, as a normal user cannot change the owner. */
-	IGNORE_CALL_RESULT(fchown(descriptor, openfile->statinfo->st_uid,
-										openfile->statinfo->st_gid));
+	 * ignore permission errors, as a normal user cannot change the owner. */
+	if (fchown(descriptor, openfile->statinfo->st_uid,
+							openfile->statinfo->st_gid) < 0 && errno != EPERM) {
+		fclose(backup_file);
+		goto problem;
+	}
 
 	/* Set the backup's permissions to those of the original file.
 	 * It is not a security issue if this fails, as we have created
 	 * the file with just read and write permission for the owner. */
-	IGNORE_CALL_RESULT(fchmod(descriptor, openfile->statinfo->st_mode));
+	if (fchmod(descriptor, openfile->statinfo->st_mode) < 0 && errno != EPERM) {
+		fclose(backup_file);
+		goto problem;
+	}
 
 	original = fopen(realname, "rb");
 
